from chalk.features import feature, features, unwrap_feature
from chalk.streams import Windowed, windowed


@features(max_staleness="1d", etl_offline_to_online=True)
class FSWithStaleness:
    parent_id: str
    scalar: int
    windowed_int: Windowed[int] = windowed("10m")

    scalar_override: int = feature(max_staleness="2d", etl_offline_to_online=False)
    windowed_int_override: Windowed[int] = windowed("10m", max_staleness="2d", etl_offline_to_online=False)


def test_assigning_max_staleness_from_class():
    """
    FeatureSet              -> FSWithStaleness
    """

    assert unwrap_feature(FSWithStaleness.scalar).max_staleness == "1d"
    assert unwrap_feature(FSWithStaleness.windowed_int).max_staleness == "1d"
    assert unwrap_feature(FSWithStaleness.windowed_int__600__).max_staleness == "1d"
    assert unwrap_feature(FSWithStaleness.scalar_override).max_staleness == "2d"
    assert unwrap_feature(FSWithStaleness.windowed_int_override).max_staleness == "2d"
    assert unwrap_feature(FSWithStaleness.windowed_int_override__600__).max_staleness == "2d"


def test_assigning_etl_offline_online_from_class():
    """
    FeatureSet              -> FSWithStaleness
    """

    assert unwrap_feature(FSWithStaleness.scalar).etl_offline_to_online == True
    assert unwrap_feature(FSWithStaleness.windowed_int).etl_offline_to_online == True
    assert unwrap_feature(FSWithStaleness.windowed_int__600__).etl_offline_to_online == True
    assert unwrap_feature(FSWithStaleness.scalar_override).etl_offline_to_online == False
    assert unwrap_feature(FSWithStaleness.windowed_int_override).etl_offline_to_online == False
    assert unwrap_feature(FSWithStaleness.windowed_int_override__600__).etl_offline_to_online == False



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, List, Mapping, Optional, Protocol, Union, overload

import pandas as pd
import polars as pl
from pydantic import BaseModel

from chalk.features import DataFrame, Feature


class OnlineQueryContext(BaseModel):
    # The environment under which to run the resolvers.
    # API tokens can be scoped to an # environment.
    # If no environment is specified in the query,
    # but the token supports only a single environment,
    # then that environment will be taken as the scope
    # for executing the request.
    environment: Optional[str] = None

    # The tags used to scope the resolvers.
    # More information at https://docs.chalk.ai/docs/resolver-tags
    tags: Optional[List[str]] = None


class OfflineQueryContext(BaseModel):
    # The environment under which to run the resolvers.
    # API tokens can be scoped to an # environment.
    # If no environment is specified in the query,
    # but the token supports only a single environment,
    # then that environment will be taken as the scope
    # for executing the request.
    environment: Optional[str] = None


class ErrorCode(str, Enum):
    # The query contained features that do not exist.
    PARSE_FAILED = "PARSE_FAILED"

    # A resolver was required as part of running the dependency
    # graph that could not be found.
    RESOLVER_NOT_FOUND = "RESOLVER_NOT_FOUND"

    # The query is invalid. All supplied features need to be
    # rooted in the same top-level entity.
    INVALID_QUERY = "INVALID_QUERY"

    # A feature value did not match the expected schema
    # (eg. `incompatible type "int"; expected "str"`)
    VALIDATION_FAILED = "VALIDATION_FAILED"

    # The resolver for a feature errored.
    RESOLVER_FAILED = "RESOLVER_FAILED"

    # A crash in a resolver that was to produce an input for
    # the resolver crashed, and so the resolver could not run
    # crashed, and so the resolver could not run.
    UPSTREAM_FAILED = "UPSTREAM_FAILED"

    # The request was submitted with an invalid authentication header.
    UNAUTHENTICATED = "UNAUTHENTICATED"

    # An unspecified error occurred.
    INTERNAL_SERVER_ERROR = "INTERNAL_SERVER_ERROR"


class ErrorCodeCategory(str, Enum):

    # Request errors are raised before execution of your
    # resolver code. They may occur due to invalid feature
    # names in the input or a request that cannot be satisfied
    # by the resolvers you have defined.
    REQUEST = "REQUEST"

    # Field errors are raised while running a feature resolver
    # for a particular field. For this type of error, you’ll
    # find a feature and resolver attribute in the error type.
    # When a feature resolver crashes, you will receive null
    # value in the response. To differentiate from a resolver
    # returning a null value and a failure in the resolver,
    # you need to check the error schema.
    FIELD = "FIELD"

    # Network errors are thrown outside your resolvers.
    # For example, your request was unauthenticated,
    # connection failed, or an error occurred within Chalk.
    NETWORK = "NETWORK"


class ChalkException(BaseModel):
    # The name of the class of the exception.
    kind: str

    # The message taken from the exception.
    message: str

    # The stacktrace produced by the code.
    stacktrace: str


class ChalkError(BaseModel):
    # The type of the error
    code: ErrorCode

    # The category of the error, given in the type field for the error codes.
    # This will be one of "REQUEST", "NETWORK", and "FIELD".
    category: ErrorCodeCategory

    # A readable description of the error message.
    message: str

    # The exception that caused the failure, if applicable.
    exception: Optional[ChalkException]

    # The fully qualified name of the failing feature, eg.user.identity.has_voip_phone.
    feature: Optional[str]

    # The fully qualified name of the failing resolver, eg.my.project.get_fraud_score.
    resolver: Optional[str]


class FeatureResult(BaseModel):
    # The name of the feature requested, eg.user.identity.has_voip_phone.
    field: str

    # The value of the requested feature.
    # If an error was encountered in resolving this feature,
    # this field will be empty.
    value: Any

    # The error code encountered in resolving this feature.
    # If no error occurred, this field is empty.
    error: Optional[ChalkError]

    # The time at which this feature was computed.
    # This value could be significantly in the past if you're using caching.
    ts: datetime


class OnlineQueryResponse(Protocol):
    # The output features and any query metadata
    data: List[FeatureResult]

    # Errors encountered while running the resolvers.
    # If no errors were encountered, this field is empty.
    errors: Optional[List[ChalkError]]

    def get_feature(self, feature: Any) -> Optional[FeatureResult]:
        """
        A convenience method for accessing feature result from the data response

        :param feature: The feature or its string representation
        :return: The FeatureResult for the feature, if it exists
        """
        ...

    def get_feature_value(self, feature: Any) -> Optional[Any]:
        """
        A convenience method for accessing feature values from the data response

        :param feature: The feature or its string representation
        :return: The value of the feature
        """
        ...


class ResolverRunStatus(str, Enum):
    RECEIVED = "received"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class ResolverRunResponse(BaseModel):
    id: str
    status: ResolverRunStatus


class WhoAmIResponse(BaseModel):
    user: str


class ChalkBaseException(Exception):
    ...


class ChalkAPIClientProtocol(Protocol):
    def trigger_resolver_run(self, resolver_fqn: str, deployment_id: Optional[str] = None) -> ResolverRunResponse:
        """
        Triggers a resolver to run.
        See https://docs.chalk.ai/docs/runs for more information.

        :param resolver_fqn: The fully qualified name of the resolver to trigger.
        :param deployment_id: Deployment ID.

        :return: Status of the resolver run and the run ID.
        """
        ...

    def get_run_status(self, run_id: str) -> ResolverRunResponse:
        """
        Retrieves the status of a resolver run.
        See https://docs.chalk.ai/docs/runs for more information.

        :param run_id: ID of the resolver run to check.

        :return: Status of the resolver run and the run ID.
        """
        ...

    def whoami(self) -> WhoAmIResponse:
        """
        Checks the identity of your client.
        Useful as a sanity test of your configuration.

        :return: the identity of your client
        """
        ...

    def query(
        self,
        input: Mapping[Union[str, Feature, Any], Any],
        output: List[Union[str, Feature, Any]],
        staleness: Optional[Mapping[Union[str, Feature, Any], str]] = None,
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> OnlineQueryResponse:
        """
        Compute features values using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information.

        :param input: The features for which there are known values, mapped to those values.
        :param output: Outputs are the features that you'd like to compute from the inputs.
        :param staleness: Maximum staleness overrides for any output features or intermediate features.
                          See https://docs.chalk.ai/docs/query-caching for more information.
        :param context: The context object controls the environment and tags
                        under which a request should execute resolvers.
        :param preview_deployment_id: If specified, Chalk will route your request to the relevant preview
                                      deployment
        :param query_name: The name for class of query you're making, for example, "loan_application_model".
        :param correlation_id: A globally unique ID for the query, used alongside logs and
                               available in web interfaces. If None, a correlation ID will be
                               generated for you and returned on the response.
        :param meta: Arbitrary key:value pairs to associate with a query.

        :return: The outputs features and any query metadata, plus errors encountered while
        running the resolvers.
        """
        ...

    def upload_features(
        self,
        input: Mapping[Union[str, Feature, Any], Any],
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> Optional[List[ChalkError]]:
        """
        Upload data to Chalk for use in offline resolvers or to prime a cache.

        :param input: The features for which there are known values, mapped to those values.
        :param context: The context object controls the environment and tags
                        under which a request should execute resolvers.
        :param preview_deployment_id: If specified, Chalk will route your request to the relevant preview
                                      deployment
        :param correlation_id: A globally unique ID for this operation, used alongside logs and
                               available in web interfaces. If None, a correlation ID will be
                               generated for you and returned on the response.
        :param query_name: Optionally associate this upload with a query name. See `.query` for more information.
        :param meta: Arbitrary key:value pairs to associate with a query.

        :return: The outputs features and any query metadata, plus errors encountered while
        running the resolvers.
        """
        ...

    @overload
    def get_training_dataframe(
        self,
        input: Mapping[Union[str, Feature], List[Any]],
        input_times: List[datetime],
        output: List[Union[str, Feature]],
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Compute feature values from the offline store.
        See https://docs.chalk.ai/docs/training-client for more information.

        :param input: The features for which there are known values, mapped to those values.
                      Each element in the list of values represents an observation in line
                      with the timestamp in `input_times`.
        :param input_times: A list of the times of the observations from `input`.
        :param output: The features that you'd like to compute from the inputs.
        :param context: The environment under which you'd like to query your data.
        :param dataset: A unique name that if provided will be used to generate and save a dataset
                        constructed from the list of features computed from the inputs
        :return: A pandas dataframe with columns equal to the names of the features in output,
                 and values representing the value of the observation for each input time.
                 The output maintains the ordering from `input`
        """
        ...

    @overload
    def get_training_dataframe(
        self,
        input: Union[pd.DataFrame, pl.DataFrame, DataFrame],
        input_times: List[datetime],
        output: List[Union[str, Feature]],
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Compute feature values from the offline store.
        See https://docs.chalk.ai/docs/training-client for more information.

        :param input: The features for which there are known values.
                      Each element in the list of values represents an observation in line
                      with the timestamp in `input_times`.
        :param input_times: A list of the times of the observations from `input`.
        :param output: The features that you'd like to compute from the inputs.
        :param context: The environment under which you'd like to query your data.
        :param dataset: A unique name that if provided will be used to generate and save a dataset
                        constructed from the list of features computed from the inputs
        :return: A dataframe with columns equal to the names of the features in output,
                 and values representing the value of the observation for each input time.
                 The output maintains the ordering from `input`
        """
        ...

    def get_training_dataframe(
        self,
        input: Union[Mapping[Union[str, Feature], Any], pl.DataFrame, pd.DataFrame, DataFrame],
        input_times: List[datetime],
        output: List[Union[str, Feature, Any]],
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Unified type signature for get_training_dataframe.
        Refer to the above two functions for information on the types.

        https://docs.chalk.ai/docs/training-client
        """
        ...

    def sample(
        self,
        output: list[Union[str, Feature, Any]],
        max_samples: Optional[int] = None,
        context: Optional[OfflineQueryContext] = None,
    ) -> pd.DataFrame:
        """
        Get the most recent feature values from the offline store.
        See https://docs.chalk.ai/docs/training-client for more information.

        :param output: The features that you'd like to sample.
        :param max_samples: The maximum number of rows to return.
        :param context: The environment under which you'd like to query your data.

        :return: A pandas dataframe with columns equal to the names of the features in output,
                 and values representing the value of the most recent observation.

        """
        ...



import json
import os
import pathlib

import tests
from chalk.importer import import_all_python_files_from_dir
from chalk.parsed.user_types_to_json import get_registered_types_as_json


def test_get_registered_types_as_json():
    """Functional test to ensure that we can import all registered types in the tests folder and convert them to JSON.
    Does not otherwise validate correctness."""
    tests_folder = pathlib.Path(os.path.abspath(os.path.dirname(tests.__file__)))
    import_all_python_files_from_dir(tests_folder)
    registered_types_json = get_registered_types_as_json(scope_to=tests_folder, failed=[])
    json.loads(registered_types_json)



import builtins
import types
from typing import Any, Dict, List, Mapping, Optional, Type


# A sentinel object to detect if a parameter is supplied or not.
# Use a class to give it a better repr.
class _MISSING_TYPE:
    def __repr__(self):
        return "<MISSING>"


MISSING = _MISSING_TYPE()


def create_fn(
    name: str,
    args: List[str],
    body: List[str],
    *,
    _globals: Optional[Dict[str, Any]] = None,
    _locals: Optional[Dict[str, Any]] = None,
    return_type: Any = MISSING,
):
    # Note that we mutate locals when exec() is called.  Caller
    # beware!  The only callers are internal to this module, so no
    # worries about external callers.
    if _locals is None:
        _locals = {}
    if "BUILTINS" not in _locals:
        _locals["BUILTINS"] = builtins
    return_annotation = ""
    if return_type is not MISSING:
        _locals["_return_type"] = return_type
        return_annotation = "->_return_type"
    body_text = "\n".join(f"  {b}" for b in body)

    # Compute the text of the entire function.
    txt = f' def {name}({",".join(args)}){return_annotation}:\n{body_text}'

    local_vars = ", ".join(_locals.keys())
    txt = f"def __create_fn__({local_vars}):\n{txt}\n return {name}"
    ns: Mapping[str, Any] = {}
    exec(txt, _globals, ns)
    return ns["__create_fn__"](**_locals)


def set_new_attribute(cls: Type, name: str, value: Any):
    # Never overwrites an existing attribute.  Returns True if the
    # attribute already exists.
    if name in cls.__dict__:
        return True
    if isinstance(value, types.FunctionType):
        value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    setattr(cls, name, value)
    return False


def field_assign(name: str, value: str, self_name: str):
    # self_name is what "self" is called in this function: don't
    # hard-code "self", since that might be a field name.
    # Only assigning values that are not "MISSING" so attempting to access
    # a missing value will immediately raise an AttributeError rather than return MISSING
    return f"if {value} is not MISSING: {self_name}.{name}={value}"



import os
import subprocess

import pytest

import chalk


@pytest.mark.skip
def test_stubgen():
    """Test that the chalkpy stubgen runs and that the generated stub file is the one that is committed."""
    stub_file_path = os.path.join(
        os.path.dirname(chalk.__file__), "..", "typings", "chalk", "features", "feature_set_decorator.pyi"
    )
    with open(stub_file_path, "r") as f:
        current_stub_file = f.read()
    subprocess.check_call(["chalkpy", "stubgen"])
    with open(stub_file_path, "r") as f:
        new_stub_file = f.read()
    assert (
        current_stub_file == new_stub_file
    ), f"The typing stubs generated via `chalkpy stubgen` differ. Please re-run this command and commit the updated stub file at {os.path.abspath(stub_file_path)}"



import pytest

from chalk import realtime
from chalk.features import DataFrame, features
from chalk.sql import TempSQLiteFileSource
from chalk.testing import _assert_frame_equal_ignore_order

sqlite = TempSQLiteFileSource()

seed_db = """
DROP TABLE IF EXISTS mappings_table;

CREATE TABLE mappings_table (
    id INT PRIMARY KEY,
    ufrom TEXT,   -- maps to `u_from` feature, implicitly
    uto TEXT,     -- maps to `u_to` feature, implicitly
    col_a TEXT,   -- maps to `column_a` feature, explicitly
    swap_b TEXT,  -- maps to `swap_c` feature, explicitly
    swap_c TEXT,  -- maps to `swap_b` feature, explicitly
    unused TEXT   -- unused
);

INSERT INTO mappings_table(id, ufrom, uto, col_a, swap_b, swap_c, unused) VALUES
  (0, 'from', 'to', 'value_a', 'value_b', 'value_c', 'unused');
"""


@pytest.fixture(scope="module")
def init_db():
    session = sqlite.raw_session()
    for statement in seed_db.split(";"):
        if statement != "":
            session.execute(statement)

    session.commit()


@features
class MappingFeatures:
    id: int
    u_from: str
    u_to: str
    column_a: str
    swap_b: str
    swap_c: str


@realtime
def resolve_mapping_features() -> DataFrame[MappingFeatures]:
    return sqlite.query_string(
        """
        SELECT * FROM mappings_table;
        """,
        fields={
            "col_a": MappingFeatures.column_a,
            "swap_b": MappingFeatures.swap_c,
            "swap_c": MappingFeatures.swap_b,
        },
    ).all()


def test_sql_resolver_implicit_mapping_rules(init_db):
    results = resolve_mapping_features()
    _assert_frame_equal_ignore_order(
        results,
        DataFrame(
            [MappingFeatures(id=0, u_from="from", u_to="to", column_a="value_a", swap_b="value_c", swap_c="value_b")]
        ),
    )



from typing import Any, List, cast

import pytest
from pydantic import BaseModel
from typing_extensions import TypeGuard

from chalk import OnlineResolver, online
from chalk.features import DataFrame, Feature, Features, feature, features, has_one, unwrap_feature
from chalk.streams import KafkaSource, Windowed, stream, windowed


@features
class GrandchildWindow:
    id: str
    grandchildfeat: Windowed[str] = windowed("10m", "20m", max_staleness="1h")


@features
class ChildWindow:
    id: str
    childfeat: Windowed[str] = windowed("10m", "20m", max_staleness="1h")
    grand: GrandchildWindow = has_one(lambda: ChildWindow.id == GrandchildWindow.id)


@features
class StreamFeaturesWindow:
    uid: str = feature(primary=True)
    scalar_feature: Windowed[str] = windowed("10m", "20m", max_staleness="1h")
    scalar_feature_2: Windowed[str] = windowed("10m", "20m", max_staleness="1h")
    basic: str

    child_id: str
    child: ChildWindow = has_one(lambda: StreamFeaturesWindow.child_id == ChildWindow.id)


@online
def with_children(
    r: StreamFeaturesWindow.scalar_feature(window="10m"),
    r1: StreamFeaturesWindow.child.childfeat["20m"],
    r2: StreamFeaturesWindow.child.grand.grandchildfeat("10m"),
) -> StreamFeaturesWindow.basic:
    print(r, r2, r1)
    return ""


def test_has_one_windows():
    assert resolver(with_children).inputs[0].root_fqn == "stream_features_window.scalar_feature__600__"
    assert resolver(with_children).inputs[1].root_fqn == "stream_features_window.child.childfeat__1200__"
    assert resolver(with_children).inputs[2].root_fqn == "stream_features_window.child.grand.grandchildfeat__600__"


class KafkaMessage(BaseModel):
    val_a: str


s = KafkaSource(bootstrap_server="bootstrap-server", topic="topic")


class TopicMessage(BaseModel):
    val_a: str
    val_a: str
    val_a: str


@stream(source=s)
def fn(
    message: List[TopicMessage],
) -> Features[StreamFeaturesWindow.uid, StreamFeaturesWindow.scalar_feature, StreamFeaturesWindow.scalar_feature_2]:
    return StreamFeaturesWindow(
        uid="123",
        scalar_feature="xyz",
        scalar_feature_2="abc",
    )


def resolver(x: Any) -> TypeGuard[OnlineResolver]:
    return cast(OnlineResolver, x)


@online
def accessing_via_brackets(r: StreamFeaturesWindow.scalar_feature["10m"]) -> StreamFeaturesWindow.basic:
    pass


def test_missing_window_error_message():
    with pytest.raises(TypeError) as e:

        @online
        def accessing_via_brackets(r: StreamFeaturesWindow.scalar_feature["40m"]) -> StreamFeaturesWindow.basic:
            pass

    assert str(e.value) == (
        "Unsupported window duration '40m' for 'stream_features_window.scalar_feature'. Durations '600s', '1200s' are supported."
    )


def test_windowed_inputs():
    assert resolver(accessing_via_brackets).inputs[0].fqn == "stream_features_window.scalar_feature__600__"
    assert resolver(with_children).inputs[0].fqn == "stream_features_window.scalar_feature__600__"


def test_windowed_durations_are_set():
    unwrapped = unwrap_feature(StreamFeaturesWindow.scalar_feature)
    assert isinstance(unwrapped, Feature)
    assert unwrapped.is_windowed
    assert not unwrapped.is_windowed_pseudofeature
    assert list(unwrapped.window_durations) == [600, 1200]
    with pytest.raises(ValueError):
        getattr(unwrapped, "window_duration")


def test_windowed_pseudofeature_duration_is_set():
    unwrapped = unwrap_feature(StreamFeaturesWindow.scalar_feature["10m"])
    assert isinstance(unwrapped, Feature)
    assert not unwrapped.is_windowed
    assert unwrapped.is_windowed_pseudofeature
    assert len(unwrapped.window_durations) == 0
    assert unwrapped.window_duration == 600


def test_resolver_requiring_all_windows():
    with pytest.raises(ValueError):

        @online
        def badboi(r: StreamFeaturesWindow.scalar_feature) -> StreamFeaturesWindow.basic:
            pass


def test_require_whole_thing():
    with pytest.raises(ValueError):

        @online
        def badboi(r: StreamFeaturesWindow.scalar_feature) -> StreamFeaturesWindow.basic:
            pass


def test_no_type_annotation_window_error():
    with pytest.raises(TypeError) as e:

        @features
        class BadWindow:
            a = windowed("10m")

    assert str(e.value) == "Windowed feature 'bad_window.a' is missing an annotation, like 'Windowed[str]'"


def test_missing_bucketes():
    with pytest.raises(TypeError) as e:

        @features
        class BadWindow:
            a: Windowed[int]

    assert (
        str(e.value)
        == "Windowed feature 'bad_window.a' is missing windows. To create a windowed feature, use 'a: Windowed[int] = windowed('10m', '30m')."
    )


def test_no_stream_features_from_online():

    with pytest.raises(TypeError):

        @online
        def online_resolver_stream() -> StreamFeaturesWindow.scalar_feature:
            pass

    with pytest.raises(TypeError):

        @online
        def online_df_stream() -> DataFrame[StreamFeaturesWindow.uid, StreamFeaturesWindow.scalar_feature]:
            pass


def test_streaming_resolvers_do_not_specify_windows():
    with pytest.raises(TypeError):

        @stream(source=s)
        def stream_resolver_scalar() -> StreamFeaturesWindow.uid["10m"]:
            pass

    with pytest.raises(TypeError):

        @stream(source=s)
        def stream_df_resolver() -> DataFrame[StreamFeaturesWindow.uid["10m"]]:
            pass



from datetime import datetime

from chalk.features import (
    DataFrame,
    Feature,
    Features,
    feature_time,
    features,
    has_many,
    has_one,
    online,
    unwrap_feature,
)


@features
class ChildFS:
    parent_id: str
    parents: "DataFrame[ParentFS]"
    single_parent: "SingleParentFS"


@features
class SingleChildFS:
    parent_id: str
    parent: "DataFrame[ParentFS]"


@features
class SingleParentFS:
    id: str
    children: DataFrame[ChildFS] = has_many(lambda: ChildFS.parent_id == SingleParentFS.id)


@features
class ParentFS:
    id: str
    children: "DataFrame[ChildFS]" = has_many(lambda: ChildFS.parent_id == ParentFS.id)
    single_child: "SingleChildFS" = has_one(lambda: ParentFS.id == SingleChildFS.parent_id)
    ts: datetime = feature_time()


@online
def get_parent_time(p: ParentFS.id) -> Features[ParentFS.ts]:
    thing = ParentFS(ts=datetime.now())
    return thing


def test_one_to_one():
    """
    FeatureSet              -> FeatureSet
    SingleParentFS.children -> ChildFS.single_parent
    """
    assert issubclass(ChildFS, Features)
    parent_from_features = next((x for x in ChildFS.features if x.is_has_one))
    parent_from_attr = unwrap_feature(ChildFS.single_parent)
    assert parent_from_features == parent_from_attr
    assert parent_from_features.name == "single_parent"
    assert parent_from_features.namespace == "child_fs"
    assert parent_from_features.is_has_one
    assert parent_from_features.typ is not None
    assert parent_from_features.typ.underlying == SingleParentFS
    assert parent_from_features.join == unwrap_feature(SingleParentFS.children).join
    children_typ = unwrap_feature(SingleParentFS.children).typ
    assert children_typ is not None
    assert issubclass(children_typ.parsed_annotation, DataFrame)
    assert len(children_typ.parsed_annotation.columns) == 2  # includes the autogenerated ts feature


def test_one_to_many():
    """
    FeatureSet            -> DataFrame
    ParentFS.single_child -> SingleChildFS.parent
    """
    parent_from_features: Feature = next((x for x in SingleChildFS.features if x.is_has_many))
    parent_from_attr = unwrap_feature(SingleChildFS.parent)
    assert parent_from_features == parent_from_attr
    assert parent_from_features.name == "parent"
    assert parent_from_features.namespace == "single_child_fs"
    assert parent_from_features.is_has_many
    assert parent_from_features.typ is not None
    assert issubclass(parent_from_features.typ.parsed_annotation, DataFrame)
    assert issubclass(parent_from_features.typ.parsed_annotation, DataFrame)  # for pylance
    assert parent_from_features.typ.parsed_annotation.references_feature_set == ParentFS
    assert parent_from_features.join == unwrap_feature(ParentFS.single_child).join


def test_many_to_one():
    """
    DataFrame               -> FeatureSet
    SingleParentFS.children -> ChildFS.single_parent
    """
    parent_from_features = next((x for x in ChildFS.features if x.is_has_one))
    parent_from_attr = unwrap_feature(ChildFS.single_parent)
    assert parent_from_features == parent_from_attr
    assert parent_from_features.name == "single_parent"
    assert parent_from_features.namespace == "child_fs"
    assert parent_from_features.is_has_one
    assert parent_from_features.typ is not None
    assert parent_from_features.typ.underlying == SingleParentFS
    assert parent_from_features.join == unwrap_feature(SingleParentFS.children).join


def test_many_to_many():
    """
    DataFrame         -> DataFrame
    ParentFS.children -> ChildFS.parents
    """
    parents_from_features: Feature = next((x for x in ChildFS.features if x.is_has_many))
    parents_from_attr = unwrap_feature(ChildFS.parents)
    assert parents_from_features == parents_from_attr
    assert parents_from_features.name == "parents"
    assert parents_from_features.namespace == "child_fs"
    assert parents_from_features.is_has_many
    assert parents_from_features.typ is not None
    assert isinstance(parents_from_features.typ.parsed_annotation, type)  # for pylance
    assert issubclass(parents_from_features.typ.parsed_annotation, DataFrame)
    assert issubclass(parents_from_features.typ.parsed_annotation, DataFrame)  # for pylance
    assert parents_from_features.typ.parsed_annotation.references_feature_set == ParentFS
    assert parents_from_features.join == unwrap_feature(ParentFS.children).join


def test_feature_to_string():
    # scalar feature
    assert str(ChildFS.parent_id) == "child_fs.parent_id"

    # ts feature
    assert str(ParentFS.ts) == "parent_fs.ts"

    # has-one feature
    assert str(ChildFS.single_parent) == "child_fs.single_parent"

    # scalar feature has-one
    assert str(ChildFS.single_parent.id) == "child_fs.single_parent.id"

    # Scalar feature has-many
    assert str(ChildFS.single_parent.children) == "child_fs.single_parent.children"

    # Has-many
    assert str(ParentFS.children) == "parent_fs.children"



from typing import Generic, TypeVar, cast

T = TypeVar("T")


class StateWrapper(Generic[T]):
    typ: T

    def __init__(self, typ: T):
        self.typ = typ


class StateMeta(type):
    def __getitem__(cls, item: T) -> T:
        return cast(T, StateWrapper(item))


# We're using this mechanism very intentionally
# instead of using __getitem__ on a metaclass
# to allow the editor to auto-complete the
# members of T. IntelliJ doesn't auto-complete
# on the return types of metaclass methods.
# Please check that editors are happy before
# changing this pattern.
State = StateMeta("State", (object,), {})
KeyedState = State



import re
from datetime import timedelta
from re import Pattern
from typing import Literal, Mapping, Optional, Tuple, Union

# Duration is used to describe time periods in
# natural langauge. To specify using natural
# language, write the count of the unit you would
# like, followed by the representation of the unit.
#
# Chalk support the following units:
# | Signifier | Meaning  |
# | --------- | -------- |
# | w         | Weeks    |
# | d         | Days     |
# | h         | Hours    |
# | m         | Minutes  |
# | s         | Seconds  |
#
# Examples
# | Signifier   | Meaning                           |
# | ----------- | --------------------------------- |
# | "10h"       | 10 hours                          |
# | "1w 2m"     | 1 week and 2 minutes              |
# | "1h 10m 2s" | 1 hour, 10 minutes, and 2 seconds |
#
# Read more at https://docs.chalk.ai/docs/duration
Duration = Union[str, timedelta]

# A schedule defined using the unix-cron
# string format (* * * * *).
# Values are given in the order below:
#
# | Field        | Values |
# | ------------ | ------ |
# | Minute       | 0-59   |
# | Hour         | 0-23   |
# | Day of Month | 1-31   |
# | Month        | 1-12   |
# | Day of Week  | 0-6    |
CronTab = str
ScheduleOptions = Optional[Union[CronTab, Duration, Literal[True]]]


def timedelta_to_duration(t: timedelta) -> Duration:
    s = ""
    if t.days > 0:
        s += f"{t.days}d"
    if t.seconds > 0:
        if len(s) > 0:
            s += " "
        s += f"{t.seconds}s"
    return s


_kwarg_to_regex: Mapping[str, Pattern] = {
    k: re.compile(f"(([0-9]+?){v})")
    for k, v in dict(
        weeks="w",
        days="d",
        hours="h",
        minutes="m",
        seconds="s",
    ).items()
}


def parse_chalk_duration(s: str) -> timedelta:
    def parse(regex: Pattern) -> Tuple[int, str]:
        matched = regex.search(s)
        if matched is None:
            return 0, ""

        return int(matched.groups()[1]), matched.groups()[0]

    parsed_values = {k: parse(unit) for k, unit in _kwarg_to_regex.items()}

    # Check for remaining unparsed input
    remainder = s
    for v in parsed_values.values():
        remainder = remainder.replace(v[1], "", 1)
    remainder = remainder.strip()

    if remainder != "":
        raise ValueError(f"Unparsed portion of duration: '{remainder}' for input: '{s}'")

    return timedelta(**{k: v[0] for k, v in parsed_values.items()})
